﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Analysis.AlgorithmBehavior {

  /// <summary>
  /// An operator that analyzes broken inheritance of building blocks.
  /// </summary>
  [Item("BrokenInteritanceSchemaAnalyzer", "An operator that analyzes broken inheritance of building blocks.")]
  [StorableClass]
  public sealed class BrokenInteritanceSchemaAnalyzer : SingleSuccessorOperator, IAnalyzer {
    private const string ResultsParameterName = "Results";
    private const string PopulationGraphResultParameterName = "PopulationGraph";
    private const string SchemataParameterName = "Schemata";
    private const string BrokenInteritanceSchemaOccurrenceName = "BrokenInteritanceSchemaOccurrenceMatrix";

    #region IAnalyzer Members
    public bool EnabledByDefault {
      get { return true; }
    }
    #endregion

    #region Parameter properties
    public ILookupParameter<ResultCollection> ResultsParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    #endregion

    #region Properties
    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }
    #endregion

    [StorableConstructor]
    private BrokenInteritanceSchemaAnalyzer(bool deserializing) : base(deserializing) { }
    private BrokenInteritanceSchemaAnalyzer(BrokenInteritanceSchemaAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public BrokenInteritanceSchemaAnalyzer()
      : base() {
      Parameters.Add(new LookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should are stored."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new BrokenInteritanceSchemaAnalyzer(this, cloner);
    }

    public override IOperation Apply() {
      var graph = (GenealogyGraph<Permutation>)Results[PopulationGraphResultParameterName].Value;
      var schemata = (ItemArray<IntArray>)Results[SchemataParameterName].Value;
      var generationZero = graph.Values.Where(x => x.Rank == 0);

      var brokenInteritanceSchemaOccurrenceMatrix = new StringMatrix(schemata.Count(), generationZero.Count());
      IList<string> brokenRowNames = new List<string>();
      IList<string> brokenColumnNames = new List<string>();
      for (int i = 0; i < schemata.Count(); i++) {
        var schema = schemata[i];
        brokenRowNames.Add(new Permutation(PermutationTypes.RelativeUndirected, schema).ToString());
        for (int j = 0; j < generationZero.Count(); j++) {
          var individual = generationZero.ElementAt(j);
          if (i == 0) brokenColumnNames.Add(((Permutation)individual.Data).ToString());
          brokenInteritanceSchemaOccurrenceMatrix[i, j] = CalculateBrokenInheritanceOccurrences(schema, individual).ToString();
        }
      }
      brokenInteritanceSchemaOccurrenceMatrix.RowNames = brokenRowNames;
      brokenInteritanceSchemaOccurrenceMatrix.ColumnNames = brokenColumnNames;
      Results.Add(new Result(BrokenInteritanceSchemaOccurrenceName, brokenInteritanceSchemaOccurrenceMatrix));

      return base.Apply();
    }

    private int CalculateBrokenInheritanceOccurrences(IntArray schema, GenealogyGraphNode individual) {
      var filteredDescendants = individual.Descendants().Where(x => IsMatch(schema, ((Permutation)x.Data).ToArray()));
      return filteredDescendants.Any() ? (int)filteredDescendants.Max(x => x.Rank) : 0;
    }

    // attention: brute force matching
    // should be replaced with KMP, BM or RK
    private bool IsMatch(IntArray pattern, int[] tour) {
      bool match = false;
      for (int i = 0; i <= tour.Length - pattern.Length && !match; i++) {
        for (int j = 0; j < pattern.Length && (match = pattern[j] == tour[i + j] || pattern[j] == -1); j++) ;
      }
      return match;
    }
  }
}
